/****************************************************************************

 ■FMReceiver
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC16F18323 Configuration Bit Settings

// CONFIG1
#pragma config FEXTOSC = OFF    // FEXTOSC External Oscillator mode Selection bits (Oscillator not enabled)
#pragma config RSTOSC = HFINT1  // Power-up default value for COSC bits (HFINTOSC (1MHz))
#pragma config CLKOUTEN = OFF   // Clock Out Enable bit (CLKOUT function is disabled; I/O or oscillator function on OSC2)
#pragma config CSWEN = ON       // Clock Switch Enable bit (Writing to NOSC and NDIV is allowed)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config MCLRE = ON       // Master Clear Enable bit (MCLR/VPP pin function is MCLR; Weak pull-up enabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable bits (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config LPBOREN = OFF    // Low-power BOR enable bit (ULPBOR disabled)
#pragma config BOREN = SLEEP    // Brown-out Reset Enable bits (Brown-out Reset enabled while running, disabled in Sleep; SBOREN is ignored)
#pragma config BORV = LOW       // Brown-out Reset Voltage selection bit (Brown-out voltage (Vbor) set to 2.45V)
#pragma config PPS1WAY = ON     // PPSLOCK bit One-Way Set Enable bit (The PPSLOCK bit can be cleared and set only once; PPS registers remain locked after one clear/set cycle)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable bit (Stack Overflow or Underflow will cause a Reset)
#pragma config DEBUG = OFF      // Debugger enable bit (Background debugger disabled)

// CONFIG3
#pragma config WRT = OFF        // User NVM self-write protection bits (Write protection off)
#pragma config LVP = ON         // Low Voltage Programming Enable bit (Low Voltage programming enabled. MCLR/VPP pin function is MCLR. MCLRE configuration bit is ignored.)

// CONFIG4
#pragma config CP = OFF         // User NVM Program Memory Code Protection bit (User NVM code protection disabled)
#pragma config CPD = OFF        // Data NVM Memory Code Protection bit (Data NVM code protection disabled)

#include <xc.h>

#include "bit_fields.h"

//=================================================================

#define ADDR_SI4708		0b0010000

#define SI4708_RST		LATC3
#define POWER_GOOD		RC4
#define SI4708_VIO		LATA5
#define PORT_RLY_R		LATC2
#define PORT_RLY_S		LATC1
#define PORT_PON		LATC0
#define PORT_SW			RA2
#define PORT_LED		LATA1	// ICSPCLKと共用

#define DELAY_256s		0b10010
#define DELAY_128s		0b10001
#define DELAY_64s		0b10000
#define DELAY_32s		0b01111
#define DELAY_16s		0b01110
#define DELAY_8s		0b01101
#define DELAY_4s		0b01100
#define DELAY_2s		0b01011
#define DELAY_1s		0b01010
#define DELAY_512ms		0b01001
#define DELAY_256ms		0b01000
#define DELAY_128ms		0b00111
#define DELAY_64ms		0b00110
#define DELAY_32ms		0b00101
#define DELAY_16ms		0b00100
#define DELAY_8ms		0b00011
#define DELAY_4ms		0b00010
#define DELAY_2ms		0b00001
#define DELAY_1ms		0b00000

//---------------------------------------------------------------

typedef enum _BOOL { FALSE = 0, TRUE } BOOL;
typedef unsigned char UINT8;
typedef unsigned short int UINT16;

//---------------------------------------------------------------

static void InitSystem(void);
static void Run(void);

static void Operation(void);
static void Standby(void);

static void LowBattery(void);
static BOOL SearchWaves(void);
static void OutputRelay(BOOL bOn);
static void Si470x_PowerUp(void);
static void Si470x_Config(void);
static void Si470x_Volume(UINT8 vol);
static void Si470x_Mute(BOOL mute);
static void Si470x_Tune(UINT16 freq);
static BOOL Si470x_IsStereo(void);
static void Si470x_PowerDown(void);
static void WriteRegs(UINT8 nLast);
static void ReadRegs(UINT8 nLast);
static void SleepDelay(UINT16 uWdtPs);

//---------------------------------------------------------------

#define CHAN_NUM	4
const UINT16 s_Freqs[CHAN_NUM] = { 886, 926, 971, 1065 }; // チャンネル周波数
//const UINT16 s_Freqs[CHAN_NUM] = { 765, 789, 802, 851 }; // 放送局

static UINT16 s_Regs[16] = { 0 };
static UINT8 s_nChanIdx = 0;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	PORTA = PORTC = 0x00;
	LATA = LATC = 0x00;
	
	// PORT_PON(LATC0)
	TRISCbits.TRISC0 = 0;
    ANSELCbits.ANSC0 = 0;

	// SI4708_RST(LATC3)
	TRISCbits.TRISC3 = 0;
    ANSELCbits.ANSC3 = 0;
	
	// POWER_GOOD(RC4)
    ANSELCbits.ANSC4 = 0;
	
	// SI4708_VIO(LATA5)
	TRISAbits.TRISA5 = 0;
    ANSELAbits.ANSA5 = 0;

	// PORT_RLY_R(LATC2)
	TRISCbits.TRISC2 = 0;
    ANSELCbits.ANSC2 = 0;
	
	// PORT_RLY_S(LATC1)
	TRISCbits.TRISC1 = 0;
    ANSELCbits.ANSC1 = 0;

	// PORT_SW(RA2) ※消費電力を抑えるためにプルアップは後で
    ANSELAbits.ANSA2 = 0;
	IOCAPbits.IOCAP2 = 1;	
	IOCANbits.IOCAN2 = 1;	
	
	// PORT_LED(LATA1) ※ICSPCLKと共用	
	TRISAbits.TRISA1 = 0;
    ANSELAbits.ANSA1 = 0;
	
	// I2C(Master mode)
	// SCL1:RC0/SDA1:RC1 → SCL1:RC5/SDA1:RA4
	SSP1CLKPPS = 0b10101; // Peripheral input is RC5
	SSP1DATPPS = 0b00100; // Peripheral input is RA4
	RC5PPS = 0b11000; // Rxy source is SCK1/SCL1
	RA4PPS = 0b11001; // Rxy source is SDO1/SDA1
    ANSELCbits.ANSC5 = 0;
    ANSELAbits.ANSA4 = 0;
	
    SSP1CON1bits.SSPM = 0b1000; // I2C Master mode
    SSP1ADD = 0x03; // 3未満は使えない 500÷16=31.25KHz

	// 割込み許可
	PEIE = 1;
	GIE = 1;
}

// ●実行ループ
static void Run(void)
{
	if (PCON0bits.nBOR)
		OutputRelay(FALSE);
	
	for(;;)
	{
		Operation();
		Standby();
	}
}

// ●通常動作状態
static void Operation(void)
{
	// クロック設定(500kHz)
	OSCFRQbits.HFFRQ = 0b000; // 1MHz
	OSCCON1bits.NOSC = 0b110; // HFINTOSC
	OSCCON1bits.NDIV = 0b0001; // 1MHz÷2

	// BORが発生した場合は電池残量なしと判断する
	if (!PCON0bits.nBOR)
	{
		PCON0bits.nBOR = 1;
		
		// 電圧不足 → 5回点滅してお知らせ
		LowBattery();
		return;
	}
	
	// I2C ON
    SSP1CON1bits.SSPEN = 1;

	// 電源ON
	PORT_PON = 1; // Supply VA and VD
	PORT_LED = 1;
	
	// Si4708初期化
	SI4708_VIO = 1;
	
	// RCLK(TG-3541CE)の発振開始最大時間は1s
	// 及び、パスコン(C3)チャージ
	SleepDelay(DELAY_1s);

	// 念のためI2C クリア
	SSP1STAT= 0b00000000;

	SI4708_RST = 1;
	SleepDelay(DELAY_1ms);

	// 受信開始
	Si470x_PowerUp();
	Si470x_Config();
	Si470x_Tune(s_Freqs[s_nChanIdx]);
	
	// 受信波が途絶えるか電池がなくなるまでループ
	BOOL bReceived = FALSE;
	
	BOOL bBatLow = FALSE;
	for(; !bBatLow; SleepDelay(DELAY_256ms))
	{
		if (!POWER_GOOD)
		{
			if (bReceived)
			{
				// フェードアウト
				int i = 0xE;
				do {
					Si470x_Volume(i);
					SleepDelay(DELAY_128ms);
				}
				while(--i >= 0);
			}

			bBatLow = TRUE;
			continue;
		}
		
		if (!SearchWaves())
			break;

		if (!bReceived)
		{
			bReceived = TRUE;
			OutputRelay(TRUE);
		}
	}

	// 出力リレーOFF
	OutputRelay(FALSE);

	// 受信停止
	Si470x_PowerDown();
	
	// Si4708 OFF ＆ RCLK停止
	SI4708_RST = 0;
	SI4708_VIO = 0;

	// 電源OFF
	PORT_PON = 0; // Remove VA and VD supplies
	PORT_LED = 0;
	
	// I2C OFF
    SSP1CON1bits.SSPEN = 0;

	// コンデンサの畜電力を抜く
	SleepDelay(DELAY_1s);

	if (bBatLow)
	{
		// 電圧不足 → 5回点滅してお知らせ
		LowBattery();
		return;
	}
	
	// OFF直後の再ON防止
	if (bReceived)
		SleepDelay(DELAY_8s);
}

// ●スタンバイ状態
static void Standby(void)
{
	// クロック設定(31kHz)
	OSCCON1bits.NOSC = 0b100; // LFINTOSC
	OSCCON1bits.NDIV = 0b0000;

	for(;;)
	{
		int i;

		// 振動センサ安定待ち
		WPUAbits.WPUA2 = 1;
		SleepDelay(DELAY_64ms);

		BOOL bSW = PORT_SW;
		i = 0;
		do {
			SleepDelay(DELAY_1ms);

			if (bSW != PORT_SW)
			{
				i = 0;
				bSW = PORT_SW;
			}
		} while (++i < 100);
		
		// 電源ON(振動)待ち
		IOCAFbits.IOCAF2 = 0;
		IOCIE = 1;

		SLEEP();

		IOCIE = 0;
		WPUAbits.WPUA2 = 0;
		
		// 電圧(パワーグッド)チェック
		BOOL bPG = TRUE;
		i = 0;
		do {
			if (!POWER_GOOD)
			{
				bPG = FALSE;
				break;
			}
			SleepDelay(DELAY_32ms);
		} while (++i < 8);

		if (!bPG)
		{
			// 電圧不足 → 5回点滅してお知らせ
			LowBattery();
			
			// 再び電源ON待ちに入る
			continue;
		}
		
		// 電圧OK → 通常動作へ
		break;
	}
}

// ●ローバッテリー 5回点滅してお知らせ
static void LowBattery(void)
{
	int i = 0;
	do {
		PORT_LED = 1;
		SleepDelay(DELAY_256ms);
		PORT_LED = 0;
		SleepDelay(DELAY_256ms);
		SleepDelay(DELAY_256ms);
		SleepDelay(DELAY_256ms);
	} while (++i < 5);
}

// ●プリセット局を受信する
static BOOL SearchWaves(void)
{
	if (!Si470x_IsStereo())
	{
		Si470x_Mute(TRUE);
		
		// 放送中のプリセット局を探す
		BOOL bFind = FALSE;
		int i = 2; // 念の為2回繰り返す
		do {
			int j = CHAN_NUM;
			do {
				// 次の局
				if (++s_nChanIdx >= CHAN_NUM)
					s_nChanIdx = 0;

				// 選局
				Si470x_Tune(s_Freqs[s_nChanIdx]);

				if (Si470x_IsStereo())
				{
					bFind = TRUE;
					break;
				}
			}
			while (--j > 0);
		}
		while (--i > 0 && !bFind);
		
		if (!bFind)
			return FALSE; // 電波が見つからない
	}
	
	// 音声出力ON
	if ((s_Regs[0x2] & SI47_DMUTE) == 0)
		Si470x_Volume(0xF);
	
	// 受信中
	return TRUE;
}

// ●出力リレー制御
static void OutputRelay(BOOL bOn)
{
	static char s_nRelay = -1;
	
	// TXS2-LT-3Vの仕様によりON時間は10ms
	// データシートでは誤り？によりセットとリセットが逆になっている
	if (bOn)
	{
		if (s_nRelay != 1)
		{
			PORT_RLY_R = 1;
			SleepDelay(DELAY_8ms);		
			SleepDelay(DELAY_2ms);		
			PORT_RLY_R = 0;
			
			s_nRelay = 1;
		}
	}
	else
	{
		if (s_nRelay != 0)
		{
			PORT_RLY_S = 1;
			SleepDelay(DELAY_8ms);		
			SleepDelay(DELAY_2ms);		
			PORT_RLY_S = 0;
			
			s_nRelay = 0;
		}
	}
}

// ●Si4709パワーアップ
static void Si470x_PowerUp(void)
{
	// Powerup the device
	s_Regs[0x2] = SI47_ENABLE;
	WriteRegs(0x2);
	
	SleepDelay(DELAY_128ms);	
	
    // Initialize shadow registers by reading from the part
    // Starts at register 0Ah and wraps back around till it reaches 09h
	ReadRegs(0x9);
	
	// Register 00h. Device ID
	// 15:12 PN[3:0] Part Number. 0x01 = Si4708/09
	// 11:0 MFGID[11:0] Manufacturer ID. 0x242

	// Register 01h. Chip ID
	// Si4708C Reset value = 0x1093 or 0x109E if ENABLE = 1
	// Si4708C Reset value = 0x1000 or 0x101E if ENABLE = 0
}

// ●Si4709パワーダウン
static void Si470x_PowerDown(void)
{
    // Set audio hi-z if desired
    s_Regs[0x7] |= SI47_AHIZEN;
    WriteRegs(0x7);

	// Set Register 4 [5:4] and [1:0] to 10b.
	s_Regs[0x4] &= ~0x33;
	s_Regs[0x4] |= 0x22;
	WriteRegs(0x4);

	// Set DISABLE=1
	s_Regs[0x2] = SI47_ENABLE | SI47_DISABLE;
	WriteRegs(0x2);

	// Wait for indication that powerdown was successful: ENABLE=0
	do {
		SleepDelay(DELAY_1ms);
		ReadRegs(0x2);
	}
	while (s_Regs[0x2] & SI47_ENABLE);
}

// ●Si4709初期設定
static void Si470x_Config(void)
{
	s_Regs[0x4] |= SI47_DE;
	s_Regs[0x5] |= SI47_BAND_JAPANW;
	s_Regs[0x5] |= SI47_SPACE_100KHZ;
	
	WriteRegs(0x5);			
}

// ●Si4709ボリューム設定
static void Si470x_Volume(UINT8 vol)
{
    s_Regs[0x2] |= SI47_DMUTE;

    s_Regs[0x5] &= ~SI47_VOLUME;
    s_Regs[0x5] |= vol & SI47_VOLUME;

    WriteRegs(0x5);
}

// ●Si4709ミュート
static void Si470x_Mute(BOOL mute)
{
    if (mute)
        s_Regs[0x2] &= ~SI47_DMUTE;
    else
        s_Regs[0x2] |= SI47_DMUTE;
	
    WriteRegs(0x2);
}

// ●Si4709選局
static void Si470x_Tune(UINT16 freq)
{
    UINT16 chan = freq - 760;

    // Write channel number to register 0x03 and set tune bit
    s_Regs[0x3] = chan | SI47_TUNE;
    WriteRegs(0x3);

    // Wait for stc bit to be set
    do {
		SleepDelay(DELAY_1ms);
		ReadRegs(0xA);
	}
    while ((s_Regs[0xA] & SI47_STC) == 0);

    // Write address 0x03 to clear tune bit
    s_Regs[0x3] &= ~SI47_TUNE;
    WriteRegs(0x3);

    // Wait for stc bit to be cleared.  This step is very important. If it
    // is ignored and another seek or tune occurs too quickly, the tuner
    // will not set the STC bit properly.
    do {
		SleepDelay(DELAY_1ms);
		ReadRegs(0xA);
	}
	while ((s_Regs[0xA] & SI47_STC) != 0);

    // Read address 0x0B to read status (optional)
	//ReadRegs(0xB);
}

// ●Si4709 ステレオかどうか
static BOOL Si470x_IsStereo(void)
{
    ReadRegs(0xA);
    return ((s_Regs[0xA] & SI47_ST) != 0);
}

// ●レジスタ書き込み(I2C) 0x02～(0x02～0x09)
static void WriteRegs(UINT8 nLast)
{
	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & W flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (ADDR_SI4708 << 1) | 0x00;
	while (!PIR1bits.SSP1IF);

	// data
	for (UINT8 i = 0x02; i <= nLast; i++)
	{
		PIR1bits.SSP1IF = 0;
		SSP1BUF = (UINT8)(s_Regs[i] >> 8);
		while (!PIR1bits.SSP1IF);

		PIR1bits.SSP1IF = 0;
		SSP1BUF = (UINT8)s_Regs[i];
		while (!PIR1bits.SSP1IF);
	}

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
}

// ●レジスタ読み込み(I2C) 0x0A～(0x0A～0x09)
static void ReadRegs(UINT8 nLast)
{
	if (nLast < 0x0A)
		nLast += 0x10;

	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & R flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (ADDR_SI4708 << 1) | 0x01;
	while (!PIR1bits.SSP1IF);

	// data
	for (UINT8 i = 0x0A; i <= nLast; i++)
	{
		UINT16 temp;
		
		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.RCEN = 1;
		while (!PIR1bits.SSP1IF);
		temp = SSP1BUF;

		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.ACKDT = 0;
		SSP1CON2bits.ACKEN = 1;
		while (!PIR1bits.SSP1IF);

		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.RCEN = 1;
		while (!PIR1bits.SSP1IF);
		temp = (temp << 8) | SSP1BUF;

		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.ACKDT = (i == nLast);
		SSP1CON2bits.ACKEN = 1;
		while (!PIR1bits.SSP1IF);
		
		s_Regs[i & 0x0F] = temp;
	}

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
}

// ●スリープして待つ
static void SleepDelay(UINT16 uWdtPs)
{
	WDTCONbits.WDTPS = uWdtPs;
	SWDTEN = 1;
	CLRWDT();
	SLEEP();
	SWDTEN = 0;
}

//-----------------------------------------------------------------

// ●割込み
void __interrupt() Interrupt(void)
{
	if (IOCIF)
	{
		IOCAFbits.IOCAF2 = 0;
	}
}
